/*********************************************************************
* FILEVIEW.C - This program displays text files in graphics mode     *
*   using the font and tab information supplied on the command line. *
*                                                                    *
* To compile: CL FILEVIEW.C                   Last Revised: 11/18/91 *
*********************************************************************/
#pragma comment(lib, "GRAPHICS.LIB")   /* Library-search record     */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <graph.h>
#include <bios.h>
#include <malloc.h>
#include <ctype.h>

/*----Directory holding Microsoft C graphics font files-------------*/
#define FONTDIR  "C:\\C600\\SOURCE\\SAMPLES\\*.FON"

/*----#define pseudofunctions---------------------------------------*/
#define pause() while ( !kbhit() ) ; getch()
#define beep()  printf( "%c", 0x07 )
#define LEADING ((int)(FI.pixheight/5))

/*----Local #define constants---------------------------------------*/
#define TRUE     1        /* Boolean true value                     */
#define FALSE    0        /* Boolean false value                    */
#define LF       0x0a     /* Line feed character                    */
#define CR       0x0d     /* Carriage return character              */
#define TAB      0x09     /* Tab stop character                     */
#define SPACE    0x20     /* Space character                        */
#define ESC      0x011b   /* Escape keystroke                       */
#define PGUP     0x4900   /* Page Up keystroke                      */
#define PGDN     0x5100   /* Page Down keystroke                    */
#define RIGHT    0x4d00   /* Right arrow keystroke                  */
#define LEFT     0x4b00   /* Left arrow keystroke                   */
#define UP       0x4800   /* Up arrow keystroke                     */
#define DOWN     0x5000   /* Down arrow keystroke                   */

#define MAXLINES 1024
#define WORKSTRSIZE 256

/*----Global variables----------------------------------------------*/
struct _fontinfo FI;         /* Holds font header information       */
struct videoconfig VC;       /* Holds video configuration data      */
FILE *infile=NULL;           /* File pointer for input file         */
long *LineOffset;            /* fseek() offset into input file      */
int BackColor;               /* Background color for screen         */
int ForeColor;               /* Foreground color for screen         */
int ShadowColor;             /* Color for text dropshadow           */
int DropShadow=FALSE;        /* Global dropshadow flag              */
int AtEnd=FALSE;             /* Global End-Of-File flag             */
int NumRows;                 /* Number of text rows screen can hold */
int LeftSide=0;              /* Current left margin value in pixels */
int TabSize=8;               /* Tab size indicator                  */
int LineNum=1;               /* Number of the current text line     */
int MaxLine=1;               /* Maximum text line read so far       */
char workstr[WORKSTRSIZE];   /* Tmp storage for string manipulation */
char *EndOfFile = "=====End-Of-File=====";  /* End-Of-File message  */

/*********************************************************************
* DeTab() - Replaces tabs characters in a string with an equivalent  *
*   number of space characters based on the value in TabSize.        *
*********************************************************************/
void DeTab( char *Text )
    {
    int i,j;
    char NewStr[WORKSTRSIZE];

    for ( i=j=0; i<strlen(Text); i++ )
        {
        switch( Text[i] )
            {
            case CR :  /* Handle end of line */
            case LF :
                i = strlen(Text);
                break;
            case TAB : /* Expand tab into space characters          */
                /*----Handle tab stop at tab position---------------*/
                if ( j % TabSize == 0 )
                    NewStr[j++] = SPACE;
                /*----Insert spaces until we reach next tab stop----*/
                while ( j % TabSize > 0 ) 
                    NewStr[j++] = SPACE;
                break;
            default :  /* All characters other than CR, LF, and TAB */
                /*----Output character and advance position---------*/
                NewStr[j++] = Text[i];
                break;
            }
        }
    NewStr[j] = '\0';
    strcpy( Text, NewStr );
    }

/*********************************************************************
* DrawText - Draws the supplied text string at coordinates (X,Y). If *
*   DropShadow is TRUE, a shadow is drawn behind the text.           *
*********************************************************************/
void DrawText( char *Text, int X, int Y )
    {
    if ( DropShadow )
        {
        _setcolor( ShadowColor );
        _moveto( X+1, Y );
        _outgtext( Text );
        _moveto( X, Y+1 );
        _outgtext( Text );
        _moveto( X+1, Y+1 );
        _outgtext( Text );
        }
    _setcolor ( ForeColor );
    _moveto( X, Y );
    _outgtext( Text );
    }

/*********************************************************************
* ShowPage - This function draws a screen of text based on the       *
*   current text line and the offset in Direction. This offset rep-  *
*   resents the number of lines forward or backward in the file that *
*   ShowPage should move to find the first line to draw.             *
*********************************************************************/
void ShowPage( int Direction )
    {
    int i;
    int Y=0;

    if ( AtEnd && Direction > 0 )
        {
        beep();
        return;
        }

    /*----Clear background by drawing a rectangle over screen-------*/
    _setcolor( BackColor );
    _rectangle( _GFILLINTERIOR, 0, 0, 
                VC.numxpixels-1, VC.numypixels-1 ); 

    /*----Calculate new line number and check against limits--------*/
    LineNum += Direction;
    if ( LineNum < 0 )     /* Bumped into the beginning of the file */
        {
        LineNum = 1;
        beep();
        }
    else if ( LineNum > MAXLINES )     /* Too many lines in file    */
        {
        LineNum -= Direction;
        _settextposition( 0, VC.numtextrows-1 );
        _outtext( "Too many lines!!!" );
        beep();
        }

    /*----Move to start of line in file-----------------------------*/
    if ( LineNum <= MaxLine )     /* We've been there before        */
        fseek( infile, LineOffset[LineNum], SEEK_SET );
    else 
        {                         /* Search forward to new position */
        fseek( infile, LineOffset[MaxLine], SEEK_SET );
        while ( MaxLine < LineNum &&
                fgets(workstr, WORKSTRSIZE, infile) )
            ++MaxLine;
        if ( MaxLine < LineNum )
            LineNum = MaxLine - NumRows + 1;
        }

    /*----Display one screen's worth of text------------------------*/
    AtEnd = FALSE;
    for ( i=0; i<NumRows; i++ )
        {
        if ( fgets( workstr, WORKSTRSIZE, infile ) )
            {
            if ( LineNum+i >= MaxLine )       /* Update offset table*/
                LineOffset[++MaxLine] = ftell( infile );
            DeTab( workstr );                 /* Replace tab stops  */
            DrawText( workstr, LeftSide, Y ); /* Display text       */
            Y += FI.pixheight + LEADING;      /* Advance line       */
            }
        else
            {
            beep();               /* Beep at end of file            */
            AtEnd = TRUE;
            break;                /* No point in drawing next line  */
            }
        }
    }

/*********************************************************************
* main() - Main body of program                                      *       
*********************************************************************/
int main( int argc, char *argv[] )
    {
    int i;              /* Counter                                  */
    char *ptr;          /* Temporary string pointer                 */
    unsigned Ch;        /* Temporary storage for keyboard input     */
    int HorizMove;      /* Distance to shift on left/right arrow key*/
    int NumFonts;       /* The number of graphics fonts registered  */
    int FontNum = -1;   /* Number of the graphics font we're using  */

    /*----Find and register the graphics fonts----------------------*/
    if ( (NumFonts = _registerfonts( FONTDIR )) <= 0 )
        {
        puts( "error: no fonts found" );
        exit( EXIT_FAILURE );
        }

    /*----Display syntax if no command-line data--------------------*/
    if ( argc <= 1 )
        {
        puts( "\nFILEVIEW SYNTAX:\n" );
        puts( "    FILEVIEW [/Fnn] [/Tnn] [/S] <FILENAME>\n" );
        puts( "    Where:  /Fnn specifies graphics font number" );
        puts( "            /Tnn specifies tab stop interval" );
        puts( "            /S means drawn text with dropshadows\n" );
        exit( EXIT_SUCCESS );
        }

    /*----Get command-line data-------------------------------------*/
    for ( i=1; i<argc; i++ )
        {
        ptr = argv[i];
        if ( ptr[0] == '/' )
            switch( toupper(ptr[1]) )
                {
                case 'F' :   /* User-supplied font number           */
                    Ch = atoi(2+ptr);
                    if ( Ch > 0 && Ch <= NumFonts )
                        FontNum = Ch;
                    break;
                case 'T' :   /* User-supplied tab  stop interval    */
                    Ch = atoi(2+ptr);
                    if ( Ch > 0 && Ch <= 79 )
                        TabSize = Ch;
                    break;
                case 'S' :   /* Use dropshadows on text             */
                    DropShadow = TRUE;
                    break;
                }
        else if ( !infile )
            if ( (infile = fopen( argv[i], "r" )) == NULL )
                {
                printf( "error: unable to open input file %s",
                        argv[i] );
                exit( EXIT_FAILURE );
                }
        }

    /*----Check for open input file---------------------------------*/
    if ( !infile )
        {
        puts( "error: must supply filename" );
        exit( EXIT_FAILURE );
        }

    /*----Build font selection string-------------------------------*/
    if ( FontNum == -1 )
        sprintf( workstr, "fh%d", 12 );
    else
        sprintf( workstr, "n%d", FontNum );

    /*----Select a font---------------------------------------------*/
    if ( _setfont( workstr ) == -1 )
        {
        puts( "error: unable to select font" );
        exit( EXIT_FAILURE );
        }

    /*----Change to graphics mode and get video configuration-------*/
    _setvideomode( _VRES16COLOR );
    _getvideoconfig( &VC );

    /*----Get the fontinfo and calculate NumRows based on height----*/
    _getfontinfo( &FI );
    NumRows = VC.numypixels / (FI.pixheight + LEADING) - 1;

    /*----Allocate memory for line offset table---------------------*/
    LineOffset = calloc( MAXLINES , sizeof(*LineOffset) );
    if ( !LineOffset )
        {
        _setvideomode( _DEFAULTMODE );
        puts( "error: insufficient memory" );
        exit( EXIT_FAILURE );
        }

    /*----Set screen colors-----------------------------------------*/
    _remappalette( ShadowColor=1, _BLACK );
    _remappalette( ForeColor=15, _BRIGHTWHITE );
    _setbkcolor( _BLUE );              /* Blue background           */
    _setcolor( ForeColor );            /* Bright white text         */

    /*----Process until end-of-file encountered---------------------*/
    HorizMove = FI.avgwidth * 8;
    Ch = 0;
    ShowPage( 0 );
    while ( Ch != ESC )
        {
        switch( Ch & 0xff00 )
            {
            case PGDN :
                 ShowPage( NumRows );
                 break;
            case PGUP :
                 ShowPage( -NumRows );
                 break;
            case RIGHT :
                 LeftSide -= HorizMove;
                 ShowPage( 0 );
                 break;
            case LEFT :
                 LeftSide += LeftSide < 0 ? HorizMove : 0;
                 ShowPage( 0 );
                 break;
            case UP :
                 ShowPage( -1 );
                 break;
            case DOWN :
                 ShowPage( 1 );
                 break;
            }
        Ch = _bios_keybrd(_KEYBRD_READ);
        }

    /*----Return to text mode and exit to DOS-----------------------*/
    _setvideomode( _DEFAULTMODE );
    exit( EXIT_SUCCESS );
    }
